import typing
from datetime import datetime, time, timedelta

from pytz import utc

from .base import ScheduledCallbackBase, local_tz


class ReoccurringScheduledCallback(ScheduledCallbackBase):

    def __init__(self, callback, *args, **kwargs):
        super().__init__(callback, *args, **kwargs)
        self._interval: timedelta

    def _calculate_next_call(self):
        self._next_base += self._interval
        self._update_run_time()

    def interval(self, interval: typing.Union[int, timedelta]) -> 'ReoccurringScheduledCallback':
        if isinstance(interval, int):
            interval = timedelta(seconds=interval)
        assert isinstance(interval, timedelta), type(interval)
        assert interval.total_seconds() > 0

        self._interval = interval
        super()._set_next_base(interval)

        self._update_run_time()
        return self


class DayOfWeekScheduledCallback(ScheduledCallbackBase):

    def __init__(self, callback, *args, **kwargs):
        super().__init__(callback, *args, **kwargs)
        self._time: time
        self._weekdays: typing.Set[int]

    def time(self, _time: typing.Union[time, datetime]) -> 'DayOfWeekScheduledCallback':
        super()._set_next_base(_time)

        self._time = _time if isinstance(_time, time) else _time.time()

        # it is possible that the current day is not in the weekdays -> find next correct day
        # it's also why we don't have to call _update_run_time here
        self._calculate_next_call(add_day=False)
        return self

    def weekdays(self, weekdays) -> 'DayOfWeekScheduledCallback':
        if weekdays == 'weekend':
            weekdays = [6, 7]
        elif weekdays == 'workday':
            weekdays = [1, 2, 3, 4, 5]
        elif weekdays == 'all':
            weekdays = [1, 2, 3, 4, 5, 6, 7]
        for k in weekdays:
            assert 1 <= k <= 7, k
        self._weekdays = weekdays
        return self

    def _calculate_next_call(self, add_day=True):

        # we have to do it like this so the dst-change works,
        # otherwise we have the wrong hour after the change
        next_utc = self._next_base + timedelta(days=1) if add_day else self._next_base
        loc = datetime.combine(next_utc.astimezone(local_tz).date(), self._time)

        while not loc.isoweekday() in self._weekdays:
            loc += timedelta(days=1)

        self._next_base = loc.astimezone(utc)
        self._update_run_time()



from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field


class StateOptionDefinition(BaseModel):
    value: str
    label: str


class CommandOptionDefinition(BaseModel):
    command: str
    label: str


class CommandDescriptionDefinition(BaseModel):
    command_options: Optional[List[CommandOptionDefinition]] = Field(alias='commandOptions')


class StateDescriptionDefinition(BaseModel):
    minimum: Optional[Union[int, float]]
    maximum: Optional[Union[int, float]]
    step: Optional[Union[int, float]]
    pattern: Optional[str]
    read_only: Optional[bool] = Field(alias='readOnly')
    options: Optional[List[StateOptionDefinition]]


class GroupFunctionDefinition(BaseModel):
    name: str
    params: Optional[List[str]]


class OpenhabItemDefinition(BaseModel):
    type: str
    name: str
    label: Optional[str]
    category: Optional[str]
    tags: List[str]
    link: str
    state: Any
    groups: List[str] = Field(alias='groupNames')
    members: List['OpenhabItemDefinition'] = []
    transformed_state: Optional[str] = Field(alias='transformedState')
    state_description: Optional[StateDescriptionDefinition] = Field(alias='stateDescription')
    command_description: Optional[CommandDescriptionDefinition] = Field(alias='commandDescription')
    metadata: Dict[str, Any] = {}
    editable: bool = True

    # Group only fields
    group_type: Optional[str] = Field(alias='groupType')
    group_function: Optional[GroupFunctionDefinition] = Field(alias='function')


OpenhabItemDefinition.update_forward_refs()



# 1. Static stuff
from .__version__ import __version__

# 2. Setup used libraries
import HABApp.__do_setup__

# 3. User configuration
import HABApp.config

# 4. Core features
import HABApp.core

# Import the rest
import HABApp.mqtt
import HABApp.openhab
import HABApp.rule
import HABApp.runtime

import HABApp.util
from HABApp.rule import Rule
from HABApp.parameters import Parameter, DictParameter

#from HABApp.runtime import Runtime
from HABApp.config import CONFIG


import typing


class AllEvents:
    pass


class ComplexEventValue:
    def __init__(self, value):
        self.value: typing.Any = value


class ValueUpdateEvent:
    def __init__(self, name=None, value=None):
        self.name: str = name
        self.value = value

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}>'


class ValueChangeEvent:
    def __init__(self, name=None, value=None, old_value=None):
        self.name: str = name
        self.value = value
        self.old_value = old_value

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}, old_value: {self.old_value}>'


class ItemNoChangeEvent:
    def __init__(self, name=None, seconds=None):
        self.name: str = name
        self.seconds: int = seconds

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, seconds: {self.seconds}>'


class ItemNoUpdateEvent:
    def __init__(self, name=None, seconds=None):
        self.name: str = name
        self.seconds: int = seconds

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, seconds: {self.seconds}>'



import typing
from typing import List, Optional

from pydantic import BaseModel


class HABAppThingPluginData(BaseModel):
    _val_name = 'ThingPlugin'

    created_link: Optional[str]
    created_ns: List[str] = []


# keep this up to date
cls_names = {k._val_name: k for k in (HABAppThingPluginData, )}


def load_habapp_meta(data: dict) -> dict:
    meta = data.setdefault('metadata', {})
    if meta.setdefault('HABApp', None) is None:
        return data

    cls = cls_names.get(meta['HABApp']['value'])    # type: typing.Union[HABAppThingPluginData]
    meta['HABApp'] = cls.parse_obj(meta['HABApp'].get('config', {}))
    return data


def get_api_vals(obj: typing.Union[HABAppThingPluginData]) -> typing.Tuple[str, dict]:
    return obj._val_name, obj.dict(exclude_defaults=True)



import logging
import typing
from itertools import chain
from pathlib import Path
from threading import Lock

import HABApp
from HABApp.core.logger import HABAppError
from HABApp.core.wrapper import ignore_exception
from . import name_from_path
from .file import CircularReferenceError, HABAppFile
from .watcher import AggregatingAsyncEventHandler

log = logging.getLogger('HABApp.files')

LOCK = Lock()

ALL: typing.Dict[str, HABAppFile] = {}
LOAD_RUNNING = False


def process(files: typing.List[Path], load_next: bool = True):
    global LOAD_RUNNING

    for file in files:
        name = name_from_path(file)

        # unload
        if not file.is_file():
            with LOCK:
                existing = ALL.pop(name, None)
            if existing is not None:
                existing.unload()
            continue

        try:
            # reload/initial load
            obj = HABAppFile.from_path(name, file)
        except Exception as e:
            HABAppError(log).add_exception(e, add_traceback=False).dump()
            # If we can not load the HABApp properties we skip the file
            continue

        with LOCK:
            ALL[name] = obj

            # find all which have this file as dependency and are not valid so it can be checked again
            for _f in filter(lambda x: not x.is_valid and name in x.properties.depends_on, ALL.values()):
                _f.is_checked = False

    if not load_next:
        return None

    # Start loading only once
    with LOCK:
        if LOAD_RUNNING:
            return None
        LOAD_RUNNING = True

    _load_next()


@ignore_exception
def file_load_ok(name: str):
    with LOCK:
        file = ALL.get(name)
    file.load_ok()

    # reload files with the property "reloads_on"
    reload = [k.path for k in filter(lambda f: f.is_loaded and name in f.properties.reloads_on, ALL.values())]
    if reload:
        process(reload, load_next=False)

    _load_next()


@ignore_exception
def file_load_failed(name: str):
    with LOCK:
        f = ALL.get(name)
    f.load_failed()

    _load_next()


def _load_next():
    global LOAD_RUNNING

    # check files for dependencies etc.
    for file in list(filter(lambda x: not x.is_checked, ALL.values())):
        try:
            file.check_properties()
        except Exception as e:
            if not isinstance(e, (CircularReferenceError, FileNotFoundError)):
                HABApp.core.wrapper.process_exception(file.check_properties, e, logger=log)

    # Load order is parameters -> openhab config files-> rules
    f_n = HABApp.core.files.file_name
    _all = sorted(ALL.keys())
    files = chain(filter(f_n.is_param, _all), filter(f_n.is_config, _all), filter(f_n.is_rule, _all))

    for name in files:
        file = ALL[name]
        if file.is_loaded or file.is_failed:
            continue

        if file.can_be_loaded():
            file.load()
            return None

    with LOCK:
        LOAD_RUNNING = False


def watch_folder(folder: Path, file_ending: str, watch_subfolders: bool = False) -> AggregatingAsyncEventHandler:
    handler = AggregatingAsyncEventHandler(folder, process, file_ending, watch_subfolders)
    HABApp.core.files.watcher.add_folder_watch(handler)
    return handler



from HABApp.openhab.connection_handler.func_async import \
    async_post_update, async_send_command, \
    async_get_items, async_get_item, async_item_exists, async_remove_item, async_create_item, \
    async_get_things, async_get_thing, async_set_thing_cfg, \
    async_remove_metadata, async_set_metadata, \
    async_get_persistence_data,\
    async_get_channel_link, async_remove_channel_link, async_channel_link_exists, async_create_channel_link



import typing
import datetime

from HABApp.core.const import MISSING
from HABApp.core.items.base_valueitem import BaseValueItem
from HABApp.openhab.interface import post_update, send_command, get_persistence_data


class OpenhabItem(BaseValueItem):
    """Base class for items which exists in OpenHAB.
    """

    def oh_send_command(self, value: typing.Any = MISSING):
        """Send a command to the openHAB item

        :param value: (optional) value to be sent. If not specified the item value will be used.
        """
        send_command(self.name, self.value if value is MISSING else value)

    def oh_post_update(self, value: typing.Any = MISSING):
        """Post an update to the openHAB item

        :param value: (optional) value to be posted. If not specified the item value will be used.
        """
        post_update(self.name, self.value if value is MISSING else value)

    def get_persistence_data(self, persistence: typing.Optional[str] = None,
                             start_time: typing.Optional[datetime.datetime] = None,
                             end_time: typing.Optional[datetime.datetime] = None):
        """Query historical data from the OpenHAB persistence service

        :param persistence: name of the persistence service (e.g. ``rrd4j``, ``mapdb``). If not set default will be used
        :param start_time: return only items which are newer than this
        :param end_time: return only items which are older than this
        """

        return get_persistence_data(
            self.name, persistence, start_time, end_time
        )




def get_default_logfile() -> str:
    return """version : 1


formatters:
  HABApp_format:
    format: '[%(asctime)s] [%(name)25s] %(levelname)8s | %(message)s'


handlers:
  # There are several Handlers available:
  # logging.handlers.RotatingFileHandler:
  #   Will rotate when the file reaches a certain size (see python logging documentation for args)
  # HABApp.core.lib.handler.MidnightRotatingFileHandler:
  #   Will wait until the file reaches a certain size and then rotate on midnight
  # More handlers:
  # https://docs.python.org/3/library/logging.handlers.html#rotatingfilehandler

  HABApp_default:
    class: HABApp.core.lib.handler.MidnightRotatingFileHandler
    filename: 'HABApp.log'
    maxBytes: 1_048_576
    backupCount: 3

    formatter: HABApp_format
    level: DEBUG

  EventFile:
    class: HABApp.core.lib.handler.MidnightRotatingFileHandler
    filename: 'events.log'
    maxBytes: 1_048_576
    backupCount: 3

    formatter: HABApp_format
    level: DEBUG

  BufferEventFile:
    class: logging.handlers.MemoryHandler
    capacity: 10
    formatter: HABApp_format
    target: EventFile
    level: DEBUG


loggers:
  HABApp:
    level: INFO
    handlers:
      - HABApp_default
    propagate: False

  HABApp.EventBus:
    level: INFO
    handlers:
      - BufferEventFile
    propagate: False

"""



from EasyCo import ConfigContainer, ConfigEntry


class Ping(ConfigContainer):
    enabled: bool = ConfigEntry(True, description='If enabled the configured item will show how long it takes to send '
                                                  'an update from HABApp and get the updated value back from openhab'
                                                  'in milliseconds')
    item: str = ConfigEntry('HABApp_Ping', description='Name of the Numberitem')
    interval: int = ConfigEntry(10, description='Seconds between two pings')


class General(ConfigContainer):
    listen_only: bool = ConfigEntry(
        False, description='If True HABApp will not change anything on the openHAB instance.'
    )
    wait_for_openhab: bool = ConfigEntry(
        True,
        description='If True HABApp will wait for items from the openHAB instance before loading any rules on startup'
    )


class Connection(ConfigContainer):
    host: str = 'localhost'
    port: int = 8080
    user: str = ''
    password: str = ''


class Openhab(ConfigContainer):
    ping = Ping()
    connection = Connection()
    general = General()



from typing import Any, Callable, Optional
import logging
from pathlib import Path
from HABApp.core.logger import log_error

import HABApp


def add_event_bus_listener(
        file_type: str,
        func_load: Optional[Callable[[str, Path], Any]],
        func_unload: Optional[Callable[[str, Path], Any]],
        logger: logging.Logger):

    func = {
        'config': HABApp.core.files.file_name.is_config,
        'rule': HABApp.core.files.file_name.is_rule,
        'param': HABApp.core.files.file_name.is_param,
    }[file_type]

    def filter_func_load(event: HABApp.core.events.habapp_events.RequestFileLoadEvent):
        if not func(event.name):
            return None

        name = event.name
        path = event.get_path()

        # Only load existing files
        if not path.is_file():
            log_error(logger, f'{file_type} file "{path}" does not exist and can not be loaded!')
            return None

        func_load(name, path)

    def filter_func_unload(event: HABApp.core.events.habapp_events.RequestFileUnloadEvent):
        if not func(event.name):
            return None
        name = event.name
        path = event.get_path()
        func_unload(name, path)

    if filter_func_unload is not None:
        HABApp.core.EventBus.add_listener(
            HABApp.core.EventBusListener(
                HABApp.core.const.topics.FILES,
                HABApp.core.WrappedFunction(filter_func_unload),
                HABApp.core.events.habapp_events.RequestFileUnloadEvent
            )
        )

    if filter_func_load is not None:
        HABApp.core.EventBus.add_listener(
            HABApp.core.EventBusListener(
                HABApp.core.const.topics.FILES,
                HABApp.core.WrappedFunction(filter_func_load),
                HABApp.core.events.habapp_events.RequestFileLoadEvent
            )
        )



import logging
import re
from typing import Dict, Iterator, List, Any

from HABApp.openhab.definitions.helpers.log_table import Table
from ._log import log

THING_ALIAS: Dict[str, str] = {
    'thing_uid': 'UID',
    'thing_type': 'thingTypeUID',
    'thing_location': 'location',
    'thing_label': 'label',
    'bridge_uid': 'bridgeUID',
    'editable': 'editable',
}

CHANNEL_ALIAS: Dict[str, str] = {
    'channel_uid': 'uid',
    'channel_type': 'channelTypeUID',
    'channel_label': 'label',
    'channel_kind': 'kind',
    # 'channel_description': 'description',
}


class BaseFilter:
    KEYS: Dict[str, str]

    def __init__(self, key: str, regex: str):
        self.key = key
        try:
            self.alias = self.KEYS[key]
        except KeyError:
            raise ValueError(f'Key {key} can not be used as a filter! Available: {", ".join(self.KEYS)}') from None

        try:
            self.search = re.compile(regex, re.IGNORECASE)
        except re.error as e:
            raise ValueError(f'Could not compile regex "{regex}": {e}') from None

    def matches(self, _dict: dict, test: bool) -> bool:
        v = _dict.get(self.alias)
        if v is None:
            return False
        m = self.search.fullmatch(v) is not None
        if m:
            log.log(
                logging.DEBUG if not test else logging.INFO,
                f'{self.key} "{self.search.pattern}" matches for '
                f'{_dict.get("UID", _dict.get("uid", "?UID NOT FOUND?"))}!'
            )
        return m

    def __repr__(self):
        return f'<{self.__class__.__name__} "{self.key}", "{self.search.pattern}">'


class ThingFilter(BaseFilter):
    KEYS = THING_ALIAS


class ChannelFilter(BaseFilter):
    KEYS = CHANNEL_ALIAS


def apply_filters(filters: List[BaseFilter], iterable: List[Dict[str, str]], test: bool) -> Iterator[Dict[str, Any]]:
    return filter(lambda n: all(map(lambda filter: filter.matches(n, test), filters)), iterable)


def log_overview(data: List[dict], aliases: Dict[str, str], heading=''):
    table = Table(heading)
    for k in aliases:
        table.add_column(k, align='<')
    for _item in data:
        table.add_dict({k: _item.get(alias, '') for k, alias in aliases.items()})
    for line in table.get_lines(sort_columns=['thing_type'] if 'thing_type' in table.columns else None):
        log.info(line)



from pathlib import Path
from typing import Dict, List
import re

from .cfg_validator import UserItem


RE_GROUP_NAMES = re.compile(r'([A-Za-z0-9-]+?)(?=[A-Z_ -])')


def _get_item_val_dict(field_fmt: Dict[str, str], item: UserItem):
    new = {}
    for k, format in field_fmt.items():
        if k in ('bracket_open', 'metadata', 'bracket_close'):
            continue
        val = item.__dict__[k]
        if isinstance(val, list):
            val = ', '.join(val)

        new[k] = format.format(val) if val else ''

    if item.link or item.metadata:
        new['bracket_open'] = '{'
        new['bracket_close'] = '}'
    else:
        new['bracket_open'] = ''
        new['bracket_close'] = ''

    if item.metadata:
        __m = []
        for k, __meta in item.metadata.items():
            __val = __meta['value']
            __cfg = __meta['config']

            _str = f'{k}={__val}' if not isinstance(__val, str) else f'{k}="{__val}"'
            if __cfg:
                __conf_strs = []
                for _k, _v in __cfg.items():
                    __conf_strs.append(f'{_k}={_v}' if not isinstance(_v, str) else f'{_k}="{_v}"')
                _str += f' [{", ".join(__conf_strs)}]'
            __m.append(_str)

        # link needs the "," so we indent properly
        if item.link:
            new['link'] += ','
        # metadata
        new['metadata'] = ', '.join(__m)
    else:
        new['metadata'] = ''

    return new


def _get_fmt_str(field_fmt: Dict[str, str], vals: List[Dict[str, str]]) -> str:
    w_dict = {}
    for k in field_fmt.keys():
        #
        #     w_dict[k] = 0
        #     continue

        width = max(map(len, map(lambda x: x[k], vals)), default=0)
        # indent to multiples of 4, if the entries are missing do not indent
        if width and k not in ('bracket_open', 'bracket_close', 'metadata'):
            add = width % 4
            if not add:
                add = 4
            width += add
        w_dict[k] = width

    ret = ''
    for k in field_fmt.keys():
        w = w_dict[k]
        if not w:
            ret += f'{{{k}:s}}'  # format crashes with with=0 so this is a different format string
            continue
        else:
            ret += f'{{{k}:{w}s}}'
    return ret


def create_items_file(path: Path, items_dict: Dict[str, UserItem]):
    # if we don't have any items we don't create an empty file
    if not items_dict:
        return None

    field_fmt = {
        'type': '{}',
        'name': '{}',
        'label': '"{}"',
        'icon': '<{}>',
        'groups': '({})',
        'tags': '[{}]',
        'bracket_open': '',
        'link': 'channel = "{}"',
        'metadata': '{}',
        'bracket_close': '',
    }

    grouped_items = {None: []}
    for _name, _item in items_dict.items():
        m = RE_GROUP_NAMES.match(_name)
        grp = grouped_items.setdefault(m.group(1) if m is not None else None, [])
        grp.append(_get_item_val_dict(field_fmt, _item))

    # aggregate single entry items to a block
    _aggr = []
    for _name, _items in grouped_items.items():
        if len(_items) <= 1 and _name is not None:
            _aggr.append(_name)
    for _name in _aggr:
        grouped_items[None].extend(grouped_items[_name])
        grouped_items.pop(_name)

    # single entry items get created at the end of file
    if None in grouped_items:
        grouped_items[None] = grouped_items.pop(None)

    lines = []
    for _name, _item_vals in grouped_items.items():
        # skip empty items
        if not _item_vals:
            continue

        fmt = _get_fmt_str(field_fmt, _item_vals)

        for _val in _item_vals:
            _l = fmt.format(**_val)
            lines.append(_l.strip() + '\n')

        # newline aber each name block
        lines.append('\n')

    with path.open(mode='w', encoding='utf-8') as file:
        file.writelines(lines)



import asyncio
from pathlib import Path

import HABApp.config
import HABApp.core
import HABApp.parameters.parameter_files
import HABApp.rule_manager
import HABApp.util
from HABApp.openhab import connection_logic as openhab_connection
from .shutdown_helper import ShutdownHelper


class Runtime:

    def __init__(self):
        self.shutdown = ShutdownHelper()

        self.config: HABApp.config.Config = None

        self.async_http: HABApp.rule.interfaces.AsyncHttpConnection = HABApp.rule.interfaces.AsyncHttpConnection()

        # OpenHAB
        self.openhab_connection: HABApp.openhab.OpenhabConnection = None

        # MQTT
        self.mqtt_connection: HABApp.mqtt.MqttConnection = None

        # Rule engine
        self.rule_manager: HABApp.rule_manager.RuleManager = None

        # Async Workers & shutdown callback
        HABApp.core.WrappedFunction._EVENT_LOOP = HABApp.core.const.loop
        self.shutdown.register_func(HABApp.core.WrappedFunction._WORKERS.shutdown)

    def startup(self, config_folder: Path):

        # Start Folder watcher!
        HABApp.core.files.watcher.start(self.shutdown)

        self.config_loader = HABApp.config.HABAppConfigLoader(config_folder)

        # MQTT
        self.mqtt_connection = HABApp.mqtt.MqttConnection(HABApp.config.CONFIG.mqtt, self.shutdown)
        self.mqtt_connection.connect()

        # openhab
        openhab_connection.setup(self.shutdown),

        # Parameter Files
        HABApp.parameters.parameter_files.setup_param_files()

        # Rule engine
        self.rule_manager = HABApp.rule_manager.RuleManager(self)
        self.rule_manager.setup()


    @HABApp.core.wrapper.log_exception
    def get_async(self):
        return asyncio.gather(
            self.async_http.create_client(),
            openhab_connection.start(),
            self.rule_manager.get_async(),
        )



import asyncio
import typing
from asyncio import Task, ensure_future, sleep, run_coroutine_threadsafe
from typing import Any, Awaitable, Callable, Optional
from HABApp.core.const import loop


class PendingFuture:
    def __init__(self, future: Callable[[], Awaitable[Any]], secs: typing.Union[int, float]):
        assert asyncio.iscoroutinefunction(future), type(future)
        assert isinstance(secs, (int, float)) and secs >= 0, f'{secs} ({type(secs)})'

        self.func: Callable[[], Awaitable[Any]] = future
        self.secs = secs
        self.task: Optional[Task] = None

        self.is_canceled: bool = False

    def cancel(self):
        self.is_canceled = True

        if self.task is not None:
            # only cancel if it is not run or canceled
            if not (self.task.done() or self.task.cancelled()):
                self.task.cancel()
            self.task = None

    def reset(self, thread_safe=False):
        if self.is_canceled:
            return None

        if self.task is not None:
            # only cancel if it is not run or canceled
            if not (self.task.done() or self.task.cancelled()):
                self.task.cancel()
            self.task = None

        if thread_safe:
            self.task = run_coroutine_threadsafe(self.__countdown(), loop)
        else:
            # todo: rename to asyncio.create_task once we go py3.7 only
            self.task = ensure_future(self.__countdown())

    async def __countdown(self):
        try:
            await sleep(self.secs)
            await self.func()
        except asyncio.CancelledError:
            pass



from pydantic import BaseModel, Extra


class RestBase(BaseModel):

    # default configuration for RestAPI models
    class Config:
        extra = Extra.forbid
        allow_population_by_field_name = True



from threading import Lock

import HABApp


class CounterItem(HABApp.core.items.Item):
    """Implements a simple thread safe counter"""

    # todo: Max Value and events when counter is 0 or has max value

    def __init__(self, name: str, initial_value: int = 0):
        """
        :param initial_value: Initial value of the counter
        """

        self.value: int = initial_value  # this gets overwritten but we provide a type hint anyway

        super().__init__(name=name, initial_value=initial_value)
        assert isinstance(initial_value, int), type(initial_value)

        self.__lock: Lock = Lock()
        self.__initial_value = initial_value

    def set_value(self, new_value) -> bool:
        assert isinstance(new_value, int), type(new_value)
        return super().set_value(new_value)

    def post_value(self, new_value):
        assert isinstance(new_value, int), type(new_value)
        super().post_value(new_value)

    def reset(self):
        """Reset value to initial value"""
        with self.__lock:
            self.post_value(self.__initial_value)
        return self.__initial_value

    def increase(self, step=1) -> int:
        """Increase value

        :param step: increase by this value, default = 1
        :return: value of the counter
        """
        assert isinstance(step, int), type(step)
        with self.__lock:
            self.post_value(self.value + step)
        return self.value

    def decrease(self, step=1) -> int:
        """Decrease value

        :param step: decrease by this value, default = 1
        :return: value of the counter
        """
        assert isinstance(step, int), type(step)
        with self.__lock:
            self.post_value(self.value - step)
        return self.value



import typing

from .base_event import OpenhabEvent

# smarthome/things/NAME/state -> 17
# openhab/things/NAME/state   -> 15
# todo: revert this once we go OH3 only
NAME_START: int = 15


class ThingStatusInfoEvent(OpenhabEvent):
    def __init__(self, name: str = '', status: str = '', detail: str = ''):
        super().__init__()

        self.name: str = name
        self.status: str = status
        self.detail: str = detail

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # smarthome/things/chromecast:chromecast:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa/status
        return cls(name=topic[NAME_START:-7], status=payload['status'], detail=payload['statusDetail'])

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, status: {self.status}, detail: {self.detail}>'


class ThingStatusInfoChangedEvent(OpenhabEvent):
    def __init__(self, name: str = '', status: str = '', detail: str = '', old_status: str = '', old_detail: str = ''):
        super().__init__()

        self.name: str = name
        self.status: str = status
        self.detail: str = detail
        self.old_status: str = old_status
        self.old_detail: str = old_detail

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # smarthome/things/chromecast:chromecast:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa/statuschanged
        name = topic[NAME_START:-14]
        new, old = payload
        return cls(
            name=name, status=new['status'], detail=new['statusDetail'],
            old_status=old['status'], old_detail=old['statusDetail']
        )

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, ' \
               f'status: {self.status}, detail: {self.detail}, ' \
               f'old_status: {self.old_status}, old_detail: {self.old_detail}>'


class ThingConfigStatusInfoEvent(OpenhabEvent):
    def __init__(self, name: str = '', messages: typing.List[typing.Dict[str, str]] = [{}]):
        super().__init__()

        self.name: str = name
        self.messages: typing.List[typing.Dict[str, str]] = messages

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # 'smarthome/things/zwave:device:controller:my_node/config/status'
        return cls(name=topic[NAME_START:-14], messages=payload['configStatusMessages'])

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, messages: {self.messages}>'


class ThingFirmwareStatusInfoEvent(OpenhabEvent):
    def __init__(self, name: str = '', status: str = ''):
        super().__init__()
        self.name: str = name
        self.status: str = status

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # 'smarthome/things/zwave:device:controller:my_node/firmware/status'
        return cls(name=topic[NAME_START:-16], status=payload['firmwareStatus'])

    def __repr__(self):
        return f'<{self.__class__.__name__} status: {self.status}>'



import asyncio
import datetime
from typing import Any, Optional, List, Dict

import HABApp
import HABApp.core
import HABApp.openhab.events
from HABApp.core.const import loop
from HABApp.core.items.base_valueitem import BaseValueItem, BaseItem
from HABApp.core.wrapper import log_exception
from HABApp.openhab.definitions.rest import OpenhabItemDefinition, OpenhabThingDefinition, ItemChannelLinkDefinition
from .func_async import async_post_update, async_send_command, async_create_item, async_get_item, async_get_thing, \
    async_set_metadata, async_remove_metadata, async_get_channel_link, async_create_channel_link, \
    async_remove_channel_link, async_channel_link_exists, \
    async_remove_item, async_item_exists, async_get_persistence_data
from .. import definitions
from ..definitions.helpers import OpenhabPersistenceData


@log_exception
def post_update(item_name: str, state: Any):
    """
    Post an update to the item

    :param item_name: item name or item
    :param state: new item state
    """
    assert isinstance(item_name, (str, BaseValueItem)), type(item_name)

    if isinstance(item_name, BaseValueItem):
        item_name = item_name.name

    asyncio.run_coroutine_threadsafe(async_post_update(item_name, state), loop)


@log_exception
def send_command(item_name: str, command):
    """
    Send the specified command to the item

    :param item_name: item name or item
    :param command: command
    """
    assert isinstance(item_name, (str, HABApp.openhab.items.base_item.BaseValueItem)), type(item_name)

    if isinstance(item_name, HABApp.openhab.items.base_item.BaseValueItem):
        item_name = item_name.name

    asyncio.run_coroutine_threadsafe(async_send_command(item_name, command), loop)


@log_exception
def create_item(item_type: str, name: str, label="", category="",
                tags: List[str] = [], groups: List[str] = [],
                group_type: str = '', group_function: str = '', group_function_params: List[str] = []):
    """Creates a new item in the OpenHAB item registry or updates an existing one

    :param item_type: item type
    :param name: item name
    :param label: item label
    :param category: item category
    :param tags: item tags
    :param groups: in which groups is the item
    :param group_type: what kind of group is it
    :param group_function: group state aggregation function
    :param group_function_params: params for group state aggregation
    :return: True if item was created/updated
    """

    def validate(_in):
        assert isinstance(_in, str), type(_in)

    # limit values to special entries and validate parameters
    if ':' in item_type:
        __type, __unit = item_type.split(':')
        assert __unit in definitions.ITEM_DIMENSIONS, \
            f'{__unit} is not a valid Openhab unit: {", ".join(definitions.ITEM_DIMENSIONS)}'
        assert __type in definitions.ITEM_TYPES, \
            f'{__type} is not a valid OpenHAB type: {", ".join(definitions.ITEM_TYPES)}'
    else:
        assert item_type in definitions.ITEM_TYPES, \
            f'{item_type} is not an OpenHAB type: {", ".join(definitions.ITEM_TYPES)}'
    assert isinstance(name, str), type(name)
    assert isinstance(label, str), type(label)
    assert isinstance(category, str), type(category)
    map(validate, tags)
    map(validate, groups)
    assert isinstance(group_type, str), type(group_type)
    assert isinstance(group_function, str), type(group_function)
    map(validate, group_function_params)

    if group_type or group_function or group_function_params:
        assert item_type == 'Group', f'Item type must be "Group"! Is: {item_type}'

        if group_function:
            assert group_function in definitions.GROUP_FUNCTIONS, \
                f'{item_type} is not a group function: {", ".join(definitions.GROUP_FUNCTIONS)}'


    fut = asyncio.run_coroutine_threadsafe(
        async_create_item(
            item_type, name,
            label=label, category=category, tags=tags, groups=groups,
            group_type=group_type, group_function=group_function, group_function_params=group_function_params
        ),
        loop
    )
    return fut.result()


def get_item(item_name: str, metadata: Optional[str] = None) -> OpenhabItemDefinition:
    """Return the complete OpenHAB item definition

    :param item_name: name of the item or item
    :param metadata: metadata to include (optional, comma separated or search expression)
    :return:
    """
    if isinstance(item_name, HABApp.openhab.items.base_item.BaseValueItem):
        item_name = item_name.name
    assert isinstance(item_name, str), type(item_name)
    assert metadata is None or isinstance(metadata, str), type(metadata)

    fut = asyncio.run_coroutine_threadsafe(async_get_item(item_name, metadata=metadata), loop)
    data = fut.result()
    return OpenhabItemDefinition.parse_obj(data)


def get_thing(thing_name: str) -> OpenhabThingDefinition:
    """ Return the complete OpenHAB thing definition

    :param thing_name: name of the thing or the item
    """
    if isinstance(thing_name, BaseItem):
        thing_name = thing_name.name
    assert isinstance(thing_name, str), type(thing_name)

    fut = asyncio.run_coroutine_threadsafe(async_get_thing(thing_name), loop)
    return fut.result()


@log_exception
def remove_item(item_name: str):
    """
    Removes an item from the openHAB item registry

    :param item_name: name
    """
    fut = asyncio.run_coroutine_threadsafe(async_remove_item(item_name), loop)
    return fut.result()


def item_exists(item_name: str):
    """
    Check if an item exists in the OpenHAB item registry

    :param item_name: name
    """
    assert isinstance(item_name, str), type(item_name)
    fut = asyncio.run_coroutine_threadsafe(async_item_exists(item_name), loop)
    return fut.result()


def set_metadata(item_name: str, namespace: str, value: str, config: dict):
    """
    Add/set metadata to an item

    :param item_name: name of the item or item
    :param namespace: namespace
    :param value: value
    :param config: configuration
    :return:
    """
    if isinstance(item_name, HABApp.openhab.items.base_item.BaseValueItem):
        item_name = item_name.name
    assert isinstance(item_name, str), type(item_name)
    assert isinstance(namespace, str), type(namespace)
    assert isinstance(value, str), type(value)
    assert isinstance(config, dict), type(config)

    fut = asyncio.run_coroutine_threadsafe(
        async_set_metadata(item=item_name, namespace=namespace, value=value, config=config), loop
    )
    return fut.result()


def remove_metadata(item_name: str, namespace: str):
    """
    Remove metadata from an item

    :param item_name: name of the item or item
    :param namespace: namespace
    :return:
    """
    if isinstance(item_name, HABApp.openhab.items.base_item.BaseValueItem):
        item_name = item_name.name
    assert isinstance(item_name, str), type(item_name)
    assert isinstance(namespace, str), type(namespace)

    fut = asyncio.run_coroutine_threadsafe(
        async_remove_metadata(item=item_name, namespace=namespace), loop
    )
    return fut.result()


def get_persistence_data(item_name: str, persistence: Optional[str],
                         start_time: Optional[datetime.datetime],
                         end_time: Optional[datetime.datetime]) -> OpenhabPersistenceData:
    """Query historical data from the OpenHAB persistence service

    :param item_name: name of the persistet item
    :param persistence: name of the persistence service (e.g. ``rrd4j``, ``mapdb``). If not set default will be used
    :param start_time: return only items which are newer than this
    :param end_time: return only items which are older than this
    """
    assert isinstance(item_name, str) and item_name, item_name
    assert isinstance(persistence, str) or persistence is None, persistence
    assert isinstance(start_time, datetime.datetime) or start_time is None, start_time
    assert isinstance(end_time, datetime.datetime) or end_time is None, end_time

    fut = asyncio.run_coroutine_threadsafe(
        async_get_persistence_data(
            item_name=item_name, persistence=persistence, start_time=start_time, end_time=end_time
        ),
        loop
    )

    ret = fut.result()
    return OpenhabPersistenceData.from_dict(ret)


# ---------------------------------------------------------------------------------------------------------------------
# Link handling is experimental
# ---------------------------------------------------------------------------------------------------------------------


def get_channel_link(channel_uid: str, item_name: str) -> ItemChannelLinkDefinition:
    """ returns the ItemChannelLinkDefinition for a link between a (things) channel and an item

    :param channel_uid: uid of the (things) channel (usually something like AAAA:BBBBB:CCCCC:DDDD:0#SOME_NAME)
    :param item_name: name of the item
    :return: an instance of ItemChannelLinkDefinition or None on error
    """

    assert isinstance(channel_uid, str), type(channel_uid)
    assert isinstance(item_name, str), type(item_name)

    fut = asyncio.run_coroutine_threadsafe(async_get_channel_link(channel_uid, item_name), loop)
    return fut.result()


def create_channel_link(channel_uid: str, item_name: str, configuration: Optional[Dict[str, Any]] = None) -> bool:
    """creates a link between a (things) channel and an item

    :param channel_uid: uid of the (thing) channel (usually something like AAAA:BBBBB:CCCCC:DDDD:0#SOME_NAME)
    :param item_name: name of the item
    :param configuration: optional configuration for the channel
    :return: true on successful creation, otherwise false
    """

    assert isinstance(channel_uid, str), type(channel_uid)
    assert isinstance(item_name, str), type(item_name)
    assert isinstance(configuration, dict), type(configuration)

    fut = asyncio.run_coroutine_threadsafe(
        async_create_channel_link(item_name=item_name, channel_uid=channel_uid, configuration=configuration),
        loop
    )
    return fut.result()


def remove_channel_link(channel_uid: str, item_name: str) -> bool:
    """ removes a link between a (things) channel and an item

    :param channel_uid: uid of the (thing) channel (usually something like AAAA:BBBBB:CCCCC:DDDD:0#SOME_NAME)
    :param item_name: name of the item
    :return: true on successful removal, otherwise false
    """

    fut = asyncio.run_coroutine_threadsafe(async_remove_channel_link(channel_uid, item_name), loop)
    return fut.result()


def channel_link_exists(channel_uid: str, item_name: str) -> bool:
    """ check if a things channel is linked to an item

    :param channel_uid: uid of the linked channel (usually something like AAAA:BBBBB:CCCCC:DDDD:0#SOME_NAME)
    :param item_name: name of the linked item
    :return: true when the link exists, otherwise false
    """

    assert isinstance(channel_uid, str), type(channel_uid)
    assert isinstance(item_name, str), type(item_name)

    fut = asyncio.run_coroutine_threadsafe(async_channel_link_exists(channel_uid, item_name), loop)
    return fut.result()



import typing
from base64 import b64encode

from HABApp.openhab.items.base_item import OpenhabItem
from ..definitions import RawValue


def _convert_bytes(data: bytes, img_type: typing.Optional[str]) -> str:
    assert isinstance(data, bytes), type(data)

    # try to automatically found out what kind of file we have
    if img_type is None:
        if data.startswith(b'\xFF\xD8\xFF'):
            img_type = 'jpeg'
        elif data.startswith(b'\x89\x50\x4E\x47'):
            img_type = 'png'
    assert img_type in ('jpeg', 'png'), f'Image type: "{img_type}", File Signature: {data[:10].hex()}'

    return f'data:image/{img_type:s};base64,{b64encode(data).decode("ascii")}'


class ImageItem(OpenhabItem):
    """ImageItem which accepts and converts the data types from OpenHAB"""

    def __init__(self, name: str, initial_value=None):
        super().__init__(name, initial_value)

        # this item is unique because we also save the image type and thus have two states
        self.image_type: typing.Optional[str] = None

    def set_value(self, new_value) -> bool:
        assert isinstance(new_value, RawValue) or new_value is None, type(new_value)

        if new_value is None:
            self.image_type = None
            return super().set_value(new_value)

        # image/png
        self.image_type = new_value.type
        if self.image_type.startswith('image/'):
            self.image_type = self.image_type[6:]
        # bytes
        return super().set_value(new_value.value)

    def oh_post_update(self, data: bytes, img_type: typing.Optional[str] = None):
        """Post an update to an openhab image with new image data. Image type is automatically detected,
        in rare cases when this does not work it can be set manually.

        :param data: image data
        :param img_type: (optional) what kind of image, ``jpeg`` or ``png``
        """
        return super().oh_post_update(_convert_bytes(data, img_type))

    def oh_send_command(self, data: bytes, img_type: typing.Optional[str] = None):
        """Send a command to an openhab image with new image data. Image type is automatically detected,
        in rare cases when this does not work it can be set manually.

        :param data: image data
        :param img_type: (optional) what kind of image, ``jpeg`` or ``png``
        """
        return super().oh_send_command(_convert_bytes(data, img_type))



import HABApp
from pathlib import Path


PREFIX_CONFIGS = 'configs'
PREFIX_PARAMS = 'params'
PREFIX_RULES = 'rules'


def name_from_path(path: Path) -> str:
    _path = path.as_posix()
    d = HABApp.config.CONFIG.directories
    folders = {PREFIX_CONFIGS: d.config.as_posix(), PREFIX_PARAMS: d.param.as_posix(), PREFIX_RULES: d.rules.as_posix()}

    for prefix, folder in folders.items():
        if _path.startswith(folder):
            return prefix + '/' + _path[len(folder) + 1:]

    raise ValueError(f'Path "{path}" is not part of the configured folders!')


def path_from_name(name: str) -> Path:
    d = HABApp.config.CONFIG.directories
    folders = {PREFIX_CONFIGS: d.config.as_posix(), PREFIX_PARAMS: d.param.as_posix(), PREFIX_RULES: d.rules.as_posix()}

    for prefix, folder in folders.items():
        if name.startswith(prefix):
            return Path(folder + '/' + name[len(prefix):])

    raise ValueError(f'Prefix not found for "{name}"!')


def is_config(name: str):
    return name.startswith(PREFIX_CONFIGS)


def is_param(name: str):
    return name.startswith(PREFIX_PARAMS)


def is_rule(name: str):
    return name.startswith(PREFIX_RULES)



import typing

from EasyCo import ConfigContainer, ConfigEntry
from voluptuous import Invalid


def MqttTopicValidator(v, msg=''):
    if isinstance(v, str):
        return [(v, 0)]

    ret = []
    for i, val in enumerate(v):
        qos = 0
        if i < len(v) - 1:
            qos = v[i + 1]

        if not isinstance(val, str) and not isinstance(val, int):
            raise Invalid(msg or "Topics must consist of int and string!")

        if not isinstance(val, str):
            continue

        if isinstance(qos, int):
            if qos not in [0, 1, 2]:
                raise Invalid(msg or ("QoS must be 0, 1 or 2"))
        else:
            qos = None

        ret.append((val, qos))
    return ret


class Connection(ConfigContainer):
    client_id: str = 'HABApp'
    host: str = ''
    port: int = 8883
    user: str = ''
    password: str = ''
    tls: bool = True
    tls_insecure: bool = False


class Subscribe(ConfigContainer):
    qos: int = ConfigEntry(default=0, description='Default QoS for subscribing')
    topics: typing.List[typing.Union[str, int]] = ConfigEntry(
        default_factory=lambda: list(('#', 0)), validator=MqttTopicValidator
    )


class Publish(ConfigContainer):
    qos: int = ConfigEntry(default=0, description='Default QoS when publishing values')
    retain: bool = ConfigEntry(default=False, description='Default retain flag when publishing values')


class General(ConfigContainer):
    listen_only: bool = ConfigEntry(
        False, description='If True HABApp will not publish any value to the broker'
    )


class Mqtt(ConfigContainer):
    connection = Connection()
    subscribe = Subscribe()
    publish = Publish()
    general = General()



from HABApp.openhab.connection_handler.func_sync import \
    post_update, send_command, \
    get_item, item_exists, remove_item, create_item, \
    get_thing, get_persistence_data,\
    remove_metadata, set_metadata, \
    get_channel_link, remove_channel_link, channel_link_exists, create_channel_link



from typing import Any, Optional, Mapping

import aiohttp

from HABApp.core.const import loop
from HABApp.core.const.json import dump_json


class AsyncHttpConnection:

    def __init__(self):
        self.__client: aiohttp.ClientSession = None

    async def create_client(self):
        self.__client = aiohttp.ClientSession(json_serialize=dump_json, loop=loop)

    def get(self, url: str, params: Optional[Mapping[str, str]] = None, **kwargs: Any)\
            -> aiohttp.client._RequestContextManager:
        """http get request

        :param url: Request URL
        :param params: Mapping, iterable of tuple of key/value pairs (e.g. dict)
                       to be sent as parameters in the query string of the new request.
                       `Params example
                       <https://docs.aiohttp.org/en/stable/client_quickstart.html#passing-parameters-in-urls>`_
        :param data: Dictionary, bytes, or file-like object to send in the body of the request
                     (optional)
        :param json: Any json compatible python object, json and data parameters could not be used at the same time.
                     (optional)
        :param kwargs: See `aiohttp request <https://docs.aiohttp.org/en/stable/client_reference.html#aiohttp.request>`_
                       for further possible kwargs
        :return: awaitable
        """
        return self.__client.get(url, params=params, **kwargs)

    def post(self, url: str, params: Optional[Mapping[str, str]] = None,
             data: Any = None, json: Any = None, **kwargs: Any) -> aiohttp.client._RequestContextManager:
        """http post request

        :param url: Request URL
        :param params: Mapping, iterable of tuple of key/value pairs (e.g. dict)
                       to be sent as parameters in the query string of the new request.
                       `Params example
                       <https://docs.aiohttp.org/en/stable/client_quickstart.html#passing-parameters-in-urls>`_
        :param data: Dictionary, bytes, or file-like object to send in the body of the request
                     (optional)
        :param json: Any json compatible python object, json and data parameters could not be used at the same time.
                     (optional)
        :param kwargs: See `aiohttp request <https://docs.aiohttp.org/en/stable/client_reference.html#aiohttp.request>`_
                       for further possible kwargs
        :return: awaitable
        """
        return self.__client.post(url, params=params, data=data, json=json, **kwargs)

    def put(self, url: str, params: Optional[Mapping[str, str]] = None,
            data: Any = None, json: Any = None, **kwargs: Any) -> aiohttp.client._RequestContextManager:
        """http put request

        :param url: Request URL
        :param params: Mapping, iterable of tuple of key/value pairs (e.g. dict)
                       to be sent as parameters in the query string of the new request.
                       `Params example
                       <https://docs.aiohttp.org/en/stable/client_quickstart.html#passing-parameters-in-urls>`_
        :param data: Dictionary, bytes, or file-like object to send in the body of the request
                     (optional)
        :param json: Any json compatible python object, json and data parameters could not be used at the same time.
                     (optional)
        :param kwargs: See `aiohttp request <https://docs.aiohttp.org/en/stable/client_reference.html#aiohttp.request>`_
                       for further possible kwargs
        :return: awaitable
        """
        return self.__client.put(url, params=params, data=data, json=json, **kwargs)

    def delete(self, url: str, params: Optional[Mapping[str, str]] = None, **kwargs: Any)\
            -> aiohttp.client._RequestContextManager:

        return self.__client.delete(url, params=params, **kwargs)

    def get_client_session(self) -> aiohttp.ClientSession:
        """Return the aiohttp
        `client session object <https://docs.aiohttp.org/en/stable/client_reference.html#client-session>`_
        for use in aiohttp libraries

        :return: session object
        """
        return self.__client



import logging
import typing
from datetime import datetime, timedelta

import HABApp
from HABApp.core.lib import PendingFuture
from .base_item_watch import BaseWatch

if typing.TYPE_CHECKING:
    from .base_item import BaseItem


TMP_DATA: typing.Dict[str, 'TmpItemData'] = {}


class TmpItemData:
    def __init__(self):
        self.ts = datetime.now()
        self.update: typing.Set[BaseWatch] = set()
        self.change: typing.Set[BaseWatch] = set()

    def add_tasks(self, update, change):
        self.ts = datetime.now()
        self.update.update(update)
        self.change.update(change)

        self.clean()

    def clean(self):
        # remove canceled
        for obj in (self.update, self.change):
            canceled = [k for k in obj if k.fut.is_canceled]
            for p in canceled:
                obj.remove(p)


def add_tmp_data(item: 'BaseItem'):
    if not item._last_update.tasks and not item._last_change.tasks:
        return None

    data = TMP_DATA.setdefault(item.name, TmpItemData())
    data.add_tasks(item._last_update.tasks, item._last_change.tasks)

    CLEANUP.reset(thread_safe=True)


def restore_tmp_data(item: 'BaseItem'):
    if item.name not in TMP_DATA:
        return None

    data = TMP_DATA.pop(item.name)
    data.clean()

    for t in data.update:
        item._last_update.tasks.append(t)
    for t in data.change:
        item._last_change.tasks.append(t)


async def clean_tmp_data():
    now = datetime.now()
    diff_max = timedelta(seconds=CLEANUP.secs)

    to_del = []
    for name, obj in TMP_DATA.items():
        diff = now - obj.ts
        if diff < diff_max:
            continue

        to_del.append(name)

        # show a warning because otherwise it's not clear what is happening
        w = HABApp.core.logger.HABAppWarning(logging.getLogger('HABApp.Item'))
        w.add(f'Item {name} has been deleted {diff.total_seconds():.1f}s ago even though it has item watchers. '
              f'If it will be added again the watchers have to be created again, too!')
        w.dump()

    for name in to_del:
        TMP_DATA.pop(name)


CLEANUP = PendingFuture(clean_tmp_data, 15)



import asyncio
import collections
import time
import typing

import HABApp
from . import BaseValueItem
from ..wrapper import ignore_exception


class AggregationItem(BaseValueItem):

    @classmethod
    def get_create_item(cls, name: str):
        """Creates a new AggregationItem in HABApp and returns it or returns the already existing one with the given name

        :param name: item name
        :return: item
        """
        assert isinstance(name, str), type(name)

        try:
            item = HABApp.core.Items.get_item(name)
        except HABApp.core.Items.ItemNotFoundException:
            item = cls(name)
            HABApp.core.Items.add_item(item)

        assert isinstance(item, cls), f'{cls} != {type(item)}'
        return item

    def __init__(self, name: str):
        super().__init__(name)
        self.__period: float = 0
        self.__aggregation_func: typing.Callable[[typing.Iterable], typing.Any] = lambda x: None

        self._ts: typing.Deque[float] = collections.deque()
        self._vals: typing.Deque[typing.Any] = collections.deque()

        self.__listener: typing.Optional[HABApp.core.EventBusListener] = None

    def aggregation_func(self, func: typing.Callable[[typing.Iterable], typing.Any]) -> 'AggregationItem':
        """Set the function which will be used to aggregate all values. E.g. ``min`` or ``max``

        :param func: The function which takes an iterator an returns an aggregated value.
                     Important: the function must be **non blocking**!
        """
        self.__aggregation_func = func
        return self

    def aggregation_period(self, period: typing.Union[float, int]) -> 'AggregationItem':
        """Set the period in which the items will be aggregated

        :param period: period in seconds
        """
        assert period > 0, period
        self.__period = period
        return self

    def aggregation_source(self, source: typing.Union[BaseValueItem, str]) -> 'AggregationItem':
        """Set the source item which changes will be aggregated

        :param item_or_name: name or Item obj
        """
        # If we already have one we cancel it
        if self.__listener is not None:
            self.__listener.cancel()
            self.__listener = None

        self.__listener = HABApp.core.EventBusListener(
            topic=source.name if isinstance(source, HABApp.core.items.BaseValueItem) else source,
            callback=HABApp.core.WrappedFunction(self._add_value, name=f'{self.name}.add_value'),
            event_type=HABApp.core.events.ValueChangeEvent
        )
        HABApp.core.EventBus.add_listener(self.__listener)
        return self

    def _on_item_remove(self):
        super()._on_item_remove()

        if self.__listener is not None:
            self.__listener.cancel()
            self.__listener = None

    async def __force_update(self):
        start = time.time()
        await asyncio.sleep(self.__period)
        sleep = time.time() - start

        # we need to sleep minimum the period, otherwise the value doesn't fall out of the interval
        # sometimes asyncio.sleep returns a little bit too early - this is what gets prevented here
        if sleep < self.__period:
            await asyncio.sleep(self.__period - sleep)

        self._aggregate()

    async def _add_value(self, event: 'HABApp.core.events.ValueChangeEvent'):
        self._ts.append(time.time())
        self._vals.append(event.value)

        # do another update when the value has fallen ouf of the period
        asyncio.ensure_future(self.__force_update())

        self._aggregate()
        return None

    @ignore_exception
    def _aggregate(self):
        # first remove entries which are too old
        now = time.time()
        while True:
            ct = len(self._ts)
            if ct <= 1:
                break

            # we keep one item from before the period because its value is valid into the period
            if (now - self._ts[1]) <= self.__period:
                break

            self._ts.popleft()
            self._vals.popleft()

        # old entries are removed -> now do the aggregation
        val = self.__aggregation_func(self._vals)
        self.post_value(val)



from HABApp.openhab.items.base_item import OpenhabItem
from HABApp.openhab.items.commands import UpDownCommand, PercentCommand
from ..definitions import UpDownValue, PercentValue


class RollershutterItem(OpenhabItem, UpDownCommand, PercentCommand):

    def set_value(self, new_value) -> bool:

        if isinstance(new_value, UpDownValue):
            new_value = 0 if new_value.up else 100
        elif isinstance(new_value, PercentValue):
            new_value = new_value.value

        assert isinstance(new_value, (int, float)) or new_value is None, new_value
        return super().set_value(new_value)

    def is_up(self) -> bool:
        return self.value <= 0

    def is_down(self) -> bool:
        return self.value >= 100

    def __str__(self):
        return self.value



import asyncio
import logging
from typing import List, Optional

from HABApp.core.const import loop
from HABApp.core.wrapper import ExceptionToHABApp

log = logging.getLogger('HABApp.openhab.plugin')


class PluginBase:
    IS_CONNECTED: bool = False

    @classmethod
    def create_plugin(cls):
        c = cls()
        assert c not in PLUGINS
        PLUGINS.append(c)
        return c

    def setup(self):
        raise NotImplementedError()

    def on_connect(self):
        raise NotImplementedError()

    def on_disconnect(self):
        raise NotImplementedError()


class OnConnectPlugin(PluginBase):
    """Plugin that runs a function on connect"""
    def __init__(self):
        super().__init__()
        self.fut: Optional[asyncio.Future] = None

    def setup(self):
        pass

    def on_connect(self):
        self.fut = asyncio.ensure_future(self.on_connect_function(), loop=loop)

    def on_disconnect(self):
        if self.fut is not None:
            self.fut.cancel()
            self.fut = None

    async def on_connect_function(self):
        raise NotImplementedError()


PLUGINS: List[PluginBase] = []


def on_connect():
    PluginBase.IS_CONNECTED = True
    for p in PLUGINS:
        with ExceptionToHABApp(log, ignore_exception=True):
            p.on_connect()


def on_disconnect():
    PluginBase.IS_CONNECTED = False
    for p in PLUGINS:
        with ExceptionToHABApp(log, ignore_exception=True):
            p.on_disconnect()


def setup_plugins():
    for p in PLUGINS:
        with ExceptionToHABApp(log, ignore_exception=True):
            p.setup()
            log.debug(f'Setup {p.__class__.__name__} complete')



import typing

try:
    from typing import Final
except ImportError:
    Final = str


INFOS: Final = 'HABApp.Infos'
WARNINGS: Final = 'HABApp.Warnings'
ERRORS: Final = 'HABApp.Errors'

FILES: Final = 'HABApp.Files'


ALL: typing.List[str] = [
    WARNINGS, ERRORS, INFOS,

    FILES
]



from .connection import setup, start

# Order matters!
from .plugin_load_items import PLUGIN_LOAD_ITEMS
from .plugin_ping import PLUGIN_PING
from .plugin_thing_overview import PLUGIN_THING_OVERVIEW
from .plugin_things import PLUGIN_MANUAL_THING_CFG



ITEM_TYPES = {
    'String', 'Number', 'Switch', 'Contact', 'Dimmer', 'Rollershutter',
    'Color', 'DateTime', 'Location', 'Player', 'Group', 'Image',
}

ITEM_DIMENSIONS = {'Length', 'Temperature', 'Pressure', 'Speed', 'Intensity', 'Angle', 'Dimensionless'}

GROUP_FUNCTIONS = {'AND', 'OR', 'NAND', 'NOR', 'AVG', 'MAX', 'MIN', 'SUM'}



from typing import Any, Dict, List, Optional

from pydantic import BaseModel


class OpenhabThingChannelDefinition(BaseModel):
    linkedItems: Optional[List[str]]
    uid: Optional[str]
    id: Optional[str]
    channelTypeUID: Optional[str]
    itemType: Optional[str]
    kind: Optional[str]
    label: Optional[str]
    description: Optional[str]
    defaultTags: Optional[List[str]]
    properties: Optional[Dict[str, Any]]
    configuration: Optional[Dict[str, Any]]


class OpenhabThingDefinition(BaseModel):
    label: Optional[str]
    bridgeUID: Optional[str]
    configuration: Dict[str, Any]
    properties: Dict[str, Any]
    UID: Optional[str]
    thingTypeUID: Optional[str]
    channels: Optional[List[OpenhabThingChannelDefinition]]
    location: Optional[str]
    statusInfo: Optional[Dict[str, str]]
    firmwareStatus: Optional[Dict[str, str]]
    editable: Optional[bool]



import asyncio
import typing
from pathlib import Path
from threading import Lock

import HABApp
from HABApp.core.wrapper import ignore_exception
from .base_watcher import BaseWatcher as __BaseWatcher

LOCK = Lock()


class AggregatingAsyncEventHandler(__BaseWatcher):
    def __init__(self, folder: Path, func: typing.Callable[[typing.List[Path]], typing.Any], file_ending: str,
                 watch_subfolders: bool = False):
        super().__init__(folder, file_ending, watch_subfolders=watch_subfolders)

        self.func = func

        self.__task: typing.Optional[asyncio.Future] = None
        self.__files: typing.Set[str] = set()

    def __execute(self):
        with LOCK:
            self.func([Path(f) for f in self.__files])
            self.__files.clear()

    @ignore_exception
    def file_changed(self, dst: str):
        # this has to be thread safe!
        with LOCK:
            self.__files.add(dst)

            # cancel already running Task
            if self.__task is not None:
                self.__task.cancel()

            # and create a new one
            self.__task = asyncio.run_coroutine_threadsafe(self.__event_waiter(), loop=HABApp.core.const.loop)

    @ignore_exception
    async def __event_waiter(self):
        try:
            # debounce time
            await asyncio.sleep(0.6)

            # trigger file event
            self.__task = None
            self.__execute()
        except asyncio.CancelledError:
            pass

    def trigger_all(self):
        files = HABApp.core.lib.list_files(self.folder, self.file_ending, self.watch_subfolders)
        self.func(files)



import logging
import typing
from datetime import datetime, timedelta

from .mode_base import BaseMode


class ValueMode(BaseMode):
    """ValueMode

    :ivar datetime.datetime last_update: Timestamp of the last update/enable of this value
    :ivar typing.Optional[datetime.timedelta] auto_disable_after: Automatically disable this mode after
                                              a given timedelta on the next recalculation
    :vartype auto_disable_func: typing.Optional[typing.Callable[[typing.Any, typing.Any], bool]]
    :ivar auto_disable_func: Function which can be used to disable this mode. Any function that accepts two
                            Arguments can be used. First arg is value with lower priority, second argument is own value.
                            Return ``True`` to disable this mode.
    :vartype calc_value_func: typing.Optional[typing.Callable[[typing.Any, typing.Any], typing.Any]]
    :ivar calc_value_func: Function to calculate the new value (e.g. ``min`` or ``max``). Any function that accepts two
                           Arguments can be used. First arg is value with lower priority, second argument is own value.
    """

    def __init__(self, name: str,
                 initial_value=None, enabled: typing.Optional[bool] = None, enable_on_value: bool = True,
                 logger: typing.Optional[logging.Logger] = None,
                 auto_disable_after=None, auto_disable_func=None,
                 calc_value_func=None):
        """

        :param name: Name of the mode
        :param initial_value: initial value of the mode
        :param enabled: initial enabled state of the mode
        :param enable_on_value: If ``True`` (default) setting a value (that is not ``None``) will also enable the mode
        :param logger: logger to log events
        :param auto_disable_after: see variables
        :param auto_disable_func: see variables
        :param calc_value_func: see variables
        """
        assert isinstance(name, str), type(name)
        super().__init__(name)

        self.__value = initial_value
        self.__enabled = False

        self.last_update: datetime = datetime.now()
        self.logger = logger

        # do not call callback for initial value
        if initial_value is not None:
            self.__enabled = True
            self.__value = initial_value
        if enabled is not None:
            assert enabled in (True, False), enabled
            self.__enabled = enabled

        self.__low_prio_value = None

        assert isinstance(enable_on_value, bool), type(enable_on_value)
        self.__enable_on_value: bool = enable_on_value

        assert isinstance(auto_disable_after, timedelta) or auto_disable_after is None, type(auto_disable_after)
        self.auto_disable_after: typing.Optional[timedelta] = auto_disable_after
        self.auto_disable_func: typing.Optional[typing.Callable[[typing.Any, typing.Any], bool]] = auto_disable_func

        self.calc_value_func: typing.Optional[typing.Callable[[typing.Any, typing.Any], typing.Any]] = calc_value_func

    @property
    def value(self):
        """Returns the current value"""
        return self.__value

    @property
    def enabled(self) -> bool:
        """Returns if the value is enabled"""
        return self.__enabled

    # we don't use the setter here because of stupid inheritance
    # https://gist.github.com/Susensio/979259559e2bebcd0273f1a95d7c1e79
    def set_value(self, value, only_on_change: bool = False):
        """Set new value and recalculate overall value. If ``enable_on_value`` is set, setting a value will also
        enable the mode.

        :param value: new value
        :param only_on_change: will set/enable the mode only if value differs or the mode is disabled
        """

        # Possibility to set the mode only on change
        if only_on_change and self.__enabled and self.__value == value:
            return None

        self.__value = value
        self.last_update = datetime.now()

        if self.__enable_on_value is True and self.__value is not None:
            self.__enabled = True

        if self.logger is not None:
            self.logger.info(f'[{"x" if self.__enabled else " "}] {self.name}: {self.__value}')

        self.parent.calculate_value()
        return None

    def set_enabled(self, value: bool, only_on_change: bool = False):
        """Enable or disable this value and recalculate overall value

        :param value: True/False
        :param only_on_change: enable only on change
        :return:
        """
        assert isinstance(value, bool), value

        # Possibility to enable/disable only on change
        if only_on_change and value == self.__enabled:
            return None

        self.__enabled = value
        self.last_update = datetime.now()

        if self.logger is not None:
            self.logger.info(f'[{"x" if self.__enabled else " "}] {self.name}')

        self.parent.calculate_value()
        return None

    def calculate_lower_priority_value(self) -> typing.Any:

        self.__low_prio_value = True

        # Trigger recalculation, this way we keep the output of MultiModeValue synchronized
        # in case some mode gets enabled/disabled/etc
        self.parent.calculate_value()

        val = self.__low_prio_value
        self.__low_prio_value = None
        return val

    def calculate_value(self, value_with_lower_priority: typing.Any) -> typing.Any:

        # helper for self.calculate_lower_priority_value
        if self.__low_prio_value is not None:
            self.__low_prio_value = value_with_lower_priority

        # so we don't spam the log if we are already disabled
        if not self.__enabled:
            return value_with_lower_priority

        # Automatically disable after certain time
        if isinstance(self.auto_disable_after, timedelta):
            if datetime.now() > self.last_update + self.auto_disable_after:
                self.__enabled = False
                self.last_update = datetime.now()
                if self.logger is not None:
                    self.logger.info(f'[{"x" if self.__enabled else " "}] {self.name} '
                                     f'(after {self.auto_disable_after})!')

        # provide user function which can disable a mode
        if self.auto_disable_func is not None:
            if self.auto_disable_func(value_with_lower_priority, self.__value) is True:
                self.__enabled = False
                self.last_update = datetime.now()
                if self.logger is not None:
                    self.logger.info(f'[{"x" if self.__enabled else " "}] {self.name} (function)')

        # check if we may have disabled this mode
        if not self.__enabled:
            return value_with_lower_priority

        if self.calc_value_func is None:
            return self.__value
        return self.calc_value_func(value_with_lower_priority, self.__value)

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.name} enabled: {self.__enabled}, value: {self.__value}>'



import logging
import threading
import typing

from HABApp.core.wrapper import log_exception
from . import EventBusListener
from .events import ComplexEventValue, ValueChangeEvent

_event_log = logging.getLogger('HABApp.EventBus')
_habapp_log = logging.getLogger('HABApp')


_LOCK = threading.Lock()


_EVENT_LISTENERS: typing.Dict[str, typing.List[EventBusListener]] = {}


@log_exception
def post_event(topic: str, event):
    assert isinstance(topic, str), type(topic)

    if not isinstance(event, str):
        event_prv = str(event)
    else:
        event_prv = event[:120] + ' ...' if len(event) > 120 else event
        event_prv = "'" + event_prv.replace('\n', '\\n') + "'"

    _event_log.info(f'{topic:>20s}: {event_prv}')

    # Sometimes we have nested data structures which we need to set the value.
    # Once the value in the item registry is updated the data structures provide no benefit thus
    # we unpack the corresponding value
    try:
        if isinstance(event.value, ComplexEventValue):
            event.value = event.value.value
        if isinstance(event, ValueChangeEvent) and isinstance(event.old_value, ComplexEventValue):
            event.old_value = event.old_value.value
    except AttributeError:
        pass

    # Notify all listeners
    for listener in _EVENT_LISTENERS.get(topic, []):
        listener.notify_listeners(event)

    return None


@log_exception
def add_listener(listener: EventBusListener):
    assert isinstance(listener, EventBusListener)

    with _LOCK:
        item_listeners = _EVENT_LISTENERS.setdefault(listener.topic, [])

        # don't add the same listener twice
        if listener in item_listeners:
            _habapp_log.warning(f'Event listener for {listener.desc()} has already been added!')
            return None

        # add listener
        item_listeners.append(listener)
        _habapp_log.debug(f'Added event listener for {listener.desc()}')
        return None


@log_exception
def remove_listener(listener: EventBusListener):
    assert isinstance(listener, EventBusListener)

    with _LOCK:
        item_listeners = _EVENT_LISTENERS.get(listener.topic, [])

        # print warning if we try to remove it twice
        if listener not in item_listeners:
            _habapp_log.warning(f'Event listener for {listener.desc()} has already been removed!')
            return None

        # remove listener
        item_listeners.remove(listener)
        _habapp_log.debug(f'Removed event listener for {listener.desc()}')


@log_exception
def remove_all_listeners():
    with _LOCK:
        _EVENT_LISTENERS.clear()



import logging
import threading
from pathlib import Path

import HABApp
from HABApp.core.files import file_load_failed, file_load_ok
from .parameters import get_parameter_file, remove_parameter_file, set_parameter_file

log = logging.getLogger('HABApp.RuleParameters')

LOCK = threading.Lock()


def setup_param_files() -> bool:
    if not HABApp.CONFIG.directories.param.is_dir():
        log.info(f'Parameter files disabled: Folder {HABApp.CONFIG.directories.param} does not exist!')
        return False

    # Add event bus listener
    HABApp.core.files.add_event_bus_listener('param', load_file, unload_file, log)

    # watch folder and load all files
    watcher = HABApp.core.files.watch_folder(HABApp.CONFIG.directories.param, '.yml', True)
    watcher.trigger_all()
    return True


def load_file(name: str, path: Path):
    with LOCK:  # serialize to get proper error messages
        try:
            with path.open(mode='r', encoding='utf-8') as file:
                data = HABApp.core.const.yml.load(file)
            if data is None:
                data = {}
            set_parameter_file(path.stem, data)
        except Exception as exc:
            e = HABApp.core.logger.HABAppError(log)
            e.add(f"Could not load parameters for {name} ({path})!")
            e.add_exception(exc, add_traceback=True)
            e.dump()

            file_load_failed(name)
            return None

    log.debug(f'Loaded params from {path.name}!')
    file_load_ok(name)


def unload_file(name: str, path: Path):
    with LOCK:  # serialize to get proper error messages
        try:
            remove_parameter_file(path.stem)
        except Exception as exc:
            e = HABApp.core.logger.HABAppError(log)
            e.add(f"Could not remove parameters for {name} ({path})!")
            e.add_exception(exc, add_traceback=True)
            e.dump()
            return None

        log.debug(f'Removed params from {path.name}!')


def save_file(file: str):
    assert isinstance(file, str), type(file)
    filename = HABApp.CONFIG.directories.param / (file + '.yml')

    with LOCK:  # serialize to get proper error messages
        log.info(f'Updated {filename}')
        with filename.open('w', encoding='utf-8') as outfile:
            HABApp.core.const.yml.dump(get_parameter_file(file), outfile)



from .base import TYPING_DATE_TIME
from .one_time_cb import OneTimeCallback
from .reoccurring_cb import DayOfWeekScheduledCallback, ReoccurringScheduledCallback
from .sun_cb import SunScheduledCallback


import logging.config

import astral as _astral
import tzlocal
import voluptuous
from EasyCo import ConfigContainer, ConfigEntry

log = logging.getLogger('HABApp.Config')


class Location(ConfigContainer):
    latitude: float = ConfigEntry(default=0.0, validator=voluptuous.Any(float, int))
    longitude: float = ConfigEntry(default=0.0, validator=voluptuous.Any(float, int))
    elevation: float = ConfigEntry(default=0.0, validator=voluptuous.Any(float, int))

    def __init__(self):
        super().__init__()

        self._astral_location: _astral.LocationInfo
        self.astral_observer: _astral.Observer

    def on_all_values_set(self):
        tz = tzlocal.get_localzone()
        tz_name = str(tz)
        log.debug(f'Local Timezone: {tz_name}')

        # unsure why we need the location in 2.1
        self._astral_location = _astral.LocationInfo(name='HABApp', )
        self._astral_location.latitude = self.latitude
        self._astral_location.longitude = self.longitude
        self._astral_location.timezone = tz_name

        self.astral_observer = self._astral_location.observer
        self.astral_observer.elevation = self.elevation



import datetime
import logging
import typing
from math import ceil, floor

from pytz import utc

import HABApp
from .base_item import BaseItem

log = logging.getLogger('HABApp')


class BaseValueItem(BaseItem):
    """Simple item

    :ivar str ~.name: Name of the item (read only)
    :ivar ~.value: Value of the item, can be anything (read only)
    :ivar ~.datetime.datetime last_change: Timestamp of the last time when the item has changed the value (read only)
    :ivar ~.datetime.datetime last_update: Timestamp of the last time when the item has updated the value (read only)
    """

    def __init__(self, name: str, initial_value=None):
        super().__init__(name)

        self.value: typing.Any = initial_value

    def set_value(self, new_value) -> bool:
        """Set a new value without creating events on the event bus

        :param new_value: new value of the item
        :return: True if state has changed
        """
        state_changed = self.value != new_value

        _now = datetime.datetime.now(tz=utc)
        if state_changed:
            self._last_change.set(_now)
        self._last_update.set(_now)

        self.value = new_value
        return state_changed

    def post_value(self, new_value) -> bool:
        """Set a new value and post appropriate events on the HABApp event bus
        (``ValueUpdateEvent``, ``ValueChangeEvent``)

        :param new_value: new value of the item
        :return: True if state has changed
        """
        old_value = self.value
        state_changed = self.set_value(new_value)

        # create events
        HABApp.core.EventBus.post_event(self._name, HABApp.core.events.ValueUpdateEvent(self._name, self.value))
        if old_value != self.value:
            HABApp.core.EventBus.post_event(
                self._name, HABApp.core.events.ValueChangeEvent(self._name, value=self.value, old_value=old_value)
            )
        return state_changed

    def get_value(self, default_value=None) -> typing.Any:
        """Return the value of the item.

        :param default_value: Return this value if the item value is None
        :return: value of the item
        """
        if self.value is None:
            return default_value
        return self.value

    def __repr__(self):
        ret = ''
        for k in ['name', 'value', 'last_change', 'last_update']:
            ret += f'{", " if ret else ""}{k}: {getattr(self, k)}'
        return f'<{self.__class__.__name__} {ret:s}>'

    # only support == and != operators by default
    # __ne__ delegates to __eq__ and inverts the result so this is not overloaded separately
    def __eq__(self, other):
        return self.value == other

    def __bool__(self):
        return bool(self.value)

    # rich comparisons
    def __lt__(self, other):
        return self.value < other

    def __le__(self, other):
        return self.value <= other

    def __ge__(self, other):
        return self.value >= other

    def __gt__(self, other):
        return self.value > other

    # https://docs.python.org/3/reference/datamodel.html#emulating-numeric-types
    # These methods are called to implement the binary arithmetic operations
    def __add__(self, other):
        if isinstance(other, BaseValueItem):
            return self.value.__add__(other.value)
        return self.value.__add__(other)

    def __sub__(self, other):
        if isinstance(other, BaseValueItem):
            return self.value.__sub__(other.value)
        return self.value.__sub__(other)

    def __mul__(self, other):
        if isinstance(other, BaseValueItem):
            return self.value.__mul__(other.value)
        return self.value.__mul__(other)

    def __matmul__(self, other):
        if isinstance(other, BaseValueItem):
            return self.value.__matmul__(other.value)
        return self.value.__matmul__(other)

    def __truediv__(self, other):
        if isinstance(other, BaseValueItem):
            return self.value.__truediv__(other.value)
        return self.value.__truediv__(other)

    def __floordiv__(self, other):
        if isinstance(other, BaseValueItem):
            return self.value.__floordiv__(other.value)
        return self.value.__floordiv__(other)

    def __mod__(self, other):
        if isinstance(other, BaseValueItem):
            return self.value.__mod__(other.value)
        return self.value.__mod__(other)

    def __divmod__(self, other):
        if isinstance(other, BaseValueItem):
            return self.value.__divmod__(other.value)
        return self.value.__divmod__(other)

    def __pow__(self, other):
        if isinstance(other, BaseValueItem):
            return self.value.__pow__(other.value)
        return self.value.__pow__(other)

    def __lshift__(self, other):
        if isinstance(other, BaseValueItem):
            return self.value.__lshift__(other.value)
        return self.value.__lshift__(other)

    def __rshift__(self, other):
        if isinstance(other, BaseValueItem):
            return self.value.__rshift__(other.value)
        return self.value.__rshift__(other)

    def __and__(self, other):
        if isinstance(other, BaseValueItem):
            return self.value.__and__(other.value)
        return self.value.__and__(other)

    def __xor__(self, other):
        if isinstance(other, BaseValueItem):
            return self.value.__xor__(other.value)
        return self.value.__xor__(other)

    def __or__(self, other):
        if isinstance(other, BaseValueItem):
            return self.value.__or__(other.value)
        return self.value.__or__(other)

    # Unary arithmetic operations (-, +, abs() and ~).
    def __neg__(self):
        return self.value.__neg__()

    def __pos__(self):
        return self.value.__pos__()

    def __abs__(self):
        return self.value.__abs__()

    def __invert__(self):
        return self.value.__invert__()

    # built-in functions complex(), int() and float().
    def __complex__(self):
        return self.value.__complex__()

    def __int__(self):
        return self.value.__int__()

    def __float__(self):
        return self.value.__float__()

    # built-in function round() and math functions trunc(), floor() and ceil().
    def __round__(self, ndigits=None):
        return self.value.__round__(ndigits)

    def __trunc__(self):
        return self.value.__trunc__()

    def __floor__(self):
        return floor(self.value)

    def __ceil__(self):
        return ceil(self.value)

    # we don't support modification in place! We have to override this because otherwise
    # python falls back to the methods above
    def __iadd__(self, other):
        return PermissionError('Call not allowed! Use "set_value" or "post_value"')

    def __isub__(self, other):
        return PermissionError('Call not allowed! Use "set_value" or "post_value"')

    def __imul__(self, other):
        return PermissionError('Call not allowed! Use "set_value" or "post_value"')

    def __imatmul__(self, other):
        return PermissionError('Call not allowed! Use "set_value" or "post_value"')

    def __itruediv__(self, other):
        return PermissionError('Call not allowed! Use "set_value" or "post_value"')

    def __ifloordiv__(self, other):
        return PermissionError('Call not allowed! Use "set_value" or "post_value"')

    def __imod__(self, other):
        return PermissionError('Call not allowed! Use "set_value" or "post_value"')

    def __ipow__(self, other):
        return PermissionError('Call not allowed! Use "set_value" or "post_value"')

    def __ilshift__(self, other):
        return PermissionError('Call not allowed! Use "set_value" or "post_value"')

    def __irshift__(self, other):
        return PermissionError('Call not allowed! Use "set_value" or "post_value"')

    def __iand__(self, other):
        return PermissionError('Call not allowed! Use "set_value" or "post_value"')

    def __ixor__(self, other):
        return PermissionError('Call not allowed! Use "set_value" or "post_value"')

    def __ior__(self, other):
        return PermissionError('Call not allowed! Use "set_value" or "post_value"')



from .base_event import OpenhabEvent

# smarthome/channels/NAME/triggered -> 19
# openhab/channels/NAME/triggered   -> 17
# todo: revert this once we go OH3 only
NAME_START: int = 17


class ChannelTriggeredEvent(OpenhabEvent):
    def __init__(self, name: str = '', event: str = '', channel: str = ''):
        super().__init__()

        self.name: str = name
        self.event: str = event
        self.channel: str = channel

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        return cls(topic[NAME_START:-10], payload['event'], payload['channel'])

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, event: {self.event}>'



import asyncio
import logging
import typing

import HABApp
from HABApp.core.lib import PendingFuture
from ..const import loop
from ..events import ItemNoChangeEvent, ItemNoUpdateEvent

log = logging.getLogger('HABApp')


class BaseWatch:
    EVENT: typing.Union[typing.Type[ItemNoUpdateEvent], typing.Type[ItemNoChangeEvent]]

    def __init__(self, name: str, secs: typing.Union[int, float]):
        self.fut = PendingFuture(self._post_event, secs)
        self.name: str = name

    async def _post_event(self):
        HABApp.core.EventBus.post_event(self.name, self.EVENT(self.name, self.fut.secs))

    async def __cancel_watch(self):
        self.fut.cancel()
        log.debug(f'Canceled {self.__class__.__name__} ({self.fut.secs}s) for {self.name}')

    def cancel(self):
        """Cancel the item watch"""
        asyncio.run_coroutine_threadsafe(self.__cancel_watch(), loop)

    def listen_event(self, callback: typing.Callable[[typing.Any], typing.Any]) -> 'HABApp.core.EventBusListener':
        """Listen to (only) the event that is emitted by this watcher"""
        rule = HABApp.rule.get_parent_rule()
        cb = HABApp.core.WrappedFunction(callback, name=rule._get_cb_name(callback))
        listener = HABApp.core.EventBusListener(
            self.name, cb, self.EVENT, 'seconds', self.fut.secs
        )
        return rule._add_event_listener(listener)


class ItemNoUpdateWatch(BaseWatch):
    EVENT = ItemNoUpdateEvent


class ItemNoChangeWatch(BaseWatch):
    EVENT = ItemNoChangeEvent



import typing
from pathlib import Path


def list_files(folder: Path, file_ending: str = '', recursive: bool = False) -> typing.List[Path]:
    # glob is much quicker than iter_dir()
    files = folder.glob(f'**/*{file_ending}' if recursive else f'*{file_ending}')
    return sorted(files, key=lambda x: x.relative_to(folder))



import HABApp
from . import BaseValueItem


class Item(BaseValueItem):
    """Simple item, used to store values in HABApp"""

    @classmethod
    def get_create_item(cls, name: str, initial_value=None):
        """Creates a new item in HABApp and returns it or returns the already existing one with the given name

        :param name: item name
        :param initial_value: state the item will have if it gets created
        :return: item
        """
        assert isinstance(name, str), type(name)

        try:
            item = HABApp.core.Items.get_item(name)
        except HABApp.core.Items.ItemNotFoundException:
            item = cls(name, initial_value)
            HABApp.core.Items.add_item(item)

        assert isinstance(item, cls), f'{cls} != {type(item)}'
        return item



import datetime

from HABApp.openhab.definitions import HSBValue, OnOffValue, OpenClosedValue, PercentValue, QuantityValue, RawValue, \
    UpDownValue


def map_openhab_values(openhab_type: str, openhab_value: str):
    assert isinstance(openhab_type, str), type(openhab_type)
    assert isinstance(openhab_value, str), type(openhab_value)

    if openhab_type == 'UnDef' or openhab_value == 'NULL':
        return None

    if openhab_type == "Number":
        return int(openhab_value)

    if openhab_type == "Decimal":
        try:
            return int(openhab_value)
        except ValueError:
            return float(openhab_value)

    if openhab_type == "DateTime":
        # 2018-11-19T09:47:38.284+0100
        dt = datetime.datetime.strptime(openhab_value.replace('+', '000+'), '%Y-%m-%dT%H:%M:%S.%f%z')
        # all datetimes from openhab have a timezone set so we can't easily compare them
        # --> TypeError: can't compare offset-naive and offset-aware datetimes
        dt = dt.astimezone(tz=None)   # Changes datetime object so it uses system timezone
        dt = dt.replace(tzinfo=None)  # Removes timezone awareness
        return dt

    if openhab_type == "HSB":
        return HSBValue(openhab_value)

    if openhab_type == 'OnOff':
        return OnOffValue(openhab_value)

    if openhab_type == 'OpenClosed':
        return OpenClosedValue(openhab_value)

    if openhab_type == 'UpDown':
        return UpDownValue(openhab_value)

    if openhab_type == 'Percent':
        return PercentValue(openhab_value)

    if openhab_type == 'Quantity':
        return QuantityValue(openhab_value)

    if openhab_type == 'Raw':
        return RawValue(openhab_value)

    return openhab_value



from datetime import datetime, timedelta

from astral import sun
from pytz import utc

import HABApp
from .base import ScheduledCallbackBase


class SunScheduledCallback(ScheduledCallbackBase):

    def __init__(self, callback, *args, **kwargs):
        super().__init__(callback, *args, **kwargs)
        self._method: str = None

    def sun_trigger(self, trig):
        assert trig in ('sunrise', 'sunset', 'dusk', 'dawn'), trig
        self._method = trig

    def _calculate_next_call(self):
        func = getattr(sun, self._method)
        observer = HABApp.config.CONFIG.location.astral_observer

        dt = datetime.now().date()
        # the datetime from astral has the proper timezone set so we don't have to do anything here
        self._next_base: datetime = func(observer=observer, date=dt).replace(microsecond=0)
        self._update_run_time()

        if self._next_call < datetime.now(tz=utc):
            self._next_base: datetime = func(observer=observer, date=dt + timedelta(days=1)).replace(microsecond=0)
            self._update_run_time()
